#include <linux/module.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/scatterlist.h>
#include <linux/crypto.h>
#include <crypto/algapi.h>
#include <crypto/scatterwalk.h>

#include "ccp-crypto.h"

#define CCP_SM4_ALG_MODE_MASK		(0xf)
#define CCP_SM4_ALG_MODE_HS_MASK	(0x10)
enum ccp_sm4_alg_mode {
	CCP_SM4_ALG_MODE_ECB = CCP_SM4_MODE_ECB,
	CCP_SM4_ALG_MODE_CBC = CCP_SM4_MODE_CBC,
	CCP_SM4_ALG_MODE_OFB = CCP_SM4_MODE_OFB,
	CCP_SM4_ALG_MODE_CFB = CCP_SM4_MODE_CFB,
	CCP_SM4_ALG_MODE_CTR = CCP_SM4_MODE_CTR,
	CCP_SM4_ALG_MODE_ECB_HS = CCP_SM4_ALG_MODE_HS_MASK | CCP_SM4_MODE_ECB,
	CCP_SM4_ALG_MODE_CBC_HS = CCP_SM4_ALG_MODE_HS_MASK | CCP_SM4_MODE_CBC,
	CCP_SM4_ALG_MODE__LAST,
};

/*
 * put the connected blocks of scatter list together
 */
static int ccp_sm4_areq_connected_sglist_merge(struct ablkcipher_request *req)
{
	struct ccp_sm4_req_ctx *rctx = ablkcipher_request_ctx(req);
	int ret;

	ret = ccp_crypto_areq_connected_sglist_merge(&rctx->src_sg,
				req->src, req->nbytes);
	if (ret)
		return ret;

	ret = ccp_crypto_areq_connected_sglist_merge(&rctx->dst_sg,
				req->dst, req->nbytes);
	if (ret) {
		kfree(rctx->src_sg);
		return ret;
	}

	return 0;
}

static int ccp_sm4_complete(struct crypto_async_request *async_req, int ret)
{
	struct ablkcipher_request *req = ablkcipher_request_cast(async_req);
	struct ccp_ctx *ctx = crypto_tfm_ctx(req->base.tfm);
	struct ccp_sm4_req_ctx *rctx = ablkcipher_request_ctx(req);

	if (ret)
		goto e_free;

	if ((ctx->u.sm4.mode & CCP_SM4_ALG_MODE_MASK) !=
					CCP_SM4_ALG_MODE_ECB) {
		memcpy(req->info, rctx->iv, SM4_BLOCK_SIZE);
		memset(rctx->iv, 0, SM4_BLOCK_SIZE);
	}

e_free:
	/* kfree(NULL) is safe */
	kfree(rctx->src_sg);
	kfree(rctx->dst_sg);
	return ret;
}

static int ccp_sm4_setkey(struct crypto_ablkcipher *tfm, const u8 *key,
			  unsigned int key_len)
{
	struct ccp_ctx *ctx = crypto_tfm_ctx(crypto_ablkcipher_tfm(tfm));

	ctx->u.sm4.key_len = 0;

	/* key_len is checked by crypto_ablkcipher_type,
	 * but key isn't checked
	 */
	if (!key)
		return -EINVAL;

	memcpy(ctx->u.sm4.key, key, SM4_KEY_SIZE);
	sg_init_one(&ctx->u.sm4.key_sg, ctx->u.sm4.key, SM4_KEY_SIZE);

	ctx->u.sm4.key_len = SM4_KEY_SIZE;

	return 0;
}

static int ccp_sm4_crypt(struct ablkcipher_request *req, bool encrypt)
{
	struct ccp_ctx *ctx = crypto_tfm_ctx(req->base.tfm);
	struct ccp_sm4_req_ctx *rctx = ablkcipher_request_ctx(req);
	struct scatterlist *iv_sg = NULL;
	struct ccp_cmd *cmd = NULL;
	enum ccp_sm4_alg_mode mode;
	enum ccp_sm4_action action;
	int ret;

	if (!ctx->u.sm4.key_len)
		return -ENOKEY;

	mode = ctx->u.sm4.mode;
	if ((mode != CCP_SM4_ALG_MODE_CTR) &&
		(mode != CCP_SM4_ALG_MODE_OFB) &&
		(mode != CCP_SM4_ALG_MODE_CFB) &&
		(req->nbytes & (SM4_BLOCK_SIZE - 1)))
		return -EINVAL;

	if ((mode & CCP_SM4_ALG_MODE_MASK) != CCP_SM4_ALG_MODE_ECB) {
		if (!req->info)
			return -EINVAL;

		memcpy(rctx->iv, req->info, SM4_BLOCK_SIZE);
		iv_sg = &rctx->iv_sg;
		sg_init_one(iv_sg, rctx->iv, SM4_BLOCK_SIZE);
	}

	rctx->src_sg = NULL;
	rctx->dst_sg = NULL;
	ret = ccp_sm4_areq_connected_sglist_merge(req);
	if (ret)
		return ret;

	cmd = &rctx->cmd;
	memset(cmd, 0, sizeof(*cmd));
	INIT_LIST_HEAD(&cmd->entry);
	action = encrypt ? CCP_SM4_ACTION_ENCRYPT : CCP_SM4_ACTION_DECRYPT;
	if (mode == CCP_SM4_ALG_MODE_CTR) {
		cmd->engine = CCP_ENGINE_SM4_CTR;
		cmd->u.sm4_ctr.action = action;
		cmd->u.sm4_ctr.size = 63;
		cmd->u.sm4_ctr.step = 1;

		cmd->u.sm4_ctr.key = &ctx->u.sm4.key_sg;
		cmd->u.sm4_ctr.key_len = SM4_KEY_SIZE;
		cmd->u.sm4_ctr.iv = iv_sg;
		cmd->u.sm4_ctr.iv_len = SM4_BLOCK_SIZE;

		cmd->u.sm4_ctr.src = rctx->src_sg;
		cmd->u.sm4_ctr.dst = rctx->dst_sg;
		cmd->u.sm4_ctr.src_len = req->nbytes;

	} else {
		cmd->engine = CCP_ENGINE_SM4;
		cmd->u.sm4.mode = mode & CCP_SM4_ALG_MODE_MASK;
		cmd->u.sm4.action = action;
		if (mode & CCP_SM4_ALG_MODE_HS_MASK)
			cmd->u.sm4.select = 1;

		cmd->u.sm4.key = &ctx->u.sm4.key_sg;
		cmd->u.sm4.key_len = SM4_KEY_SIZE;
		cmd->u.sm4.iv = iv_sg;
		cmd->u.sm4.iv_len = iv_sg ? SM4_BLOCK_SIZE : 0;

		cmd->u.sm4.src = rctx->src_sg;
		cmd->u.sm4.dst = rctx->dst_sg;
		cmd->u.sm4.src_len = req->nbytes;
	}

	ret = ccp_crypto_enqueue_request(&req->base, &rctx->cmd);

	return ret;
}

static int ccp_sm4_encrypt(struct ablkcipher_request *req)
{
	return ccp_sm4_crypt(req, true);
}

static int ccp_sm4_decrypt(struct ablkcipher_request *req)
{
	return ccp_sm4_crypt(req, false);
}

static int ccp_sm4_cra_init(struct crypto_tfm *tfm)
{
	struct ccp_ctx *ctx = crypto_tfm_ctx(tfm);
	struct ccp_crypto_ablkcipher_alg *alg = ccp_crypto_ablkcipher_alg(tfm);

	ctx->complete = ccp_sm4_complete;
	ctx->u.sm4.mode = alg->mode;
	tfm->crt_ablkcipher.reqsize = sizeof(struct ccp_sm4_req_ctx);

	return 0;
}

static void ccp_sm4_cra_exit(struct crypto_tfm *tfm)
{
}

static struct crypto_alg ccp_sm4_defaults = {
	.cra_flags	= CRYPTO_ALG_TYPE_ABLKCIPHER |
			  CRYPTO_ALG_ASYNC |
			  CRYPTO_ALG_KERN_DRIVER_ONLY,
	.cra_ctxsize	= sizeof(struct ccp_ctx),
	.cra_priority	= CCP_CRA_PRIORITY,
	.cra_type	= &crypto_ablkcipher_type,
	.cra_init	= ccp_sm4_cra_init,
	.cra_exit	= ccp_sm4_cra_exit,
	.cra_module	= THIS_MODULE,
	.cra_ablkcipher	= {
		.setkey		= ccp_sm4_setkey,
		.encrypt	= ccp_sm4_encrypt,
		.decrypt	= ccp_sm4_decrypt,
		.min_keysize	= SM4_KEY_SIZE,
		.max_keysize	= SM4_KEY_SIZE,
	},
};

struct ccp_sm4_def {
	enum ccp_sm4_alg_mode mode;
	unsigned int version;
	const char *name;
	const char *driver_name;
	unsigned int blocksize;
	unsigned int ivsize;
	struct crypto_alg *alg_defaults;
};

static struct ccp_sm4_def sm4_algs[] = {
	{
		.mode		= CCP_SM4_ALG_MODE_ECB,
		.version	= CCP_VERSION(5, 0),
		.name		= "ecb(sm4)",
		.driver_name	= "ecb-sm4-ccp",
		.blocksize	= SM4_BLOCK_SIZE,
		.ivsize		= 0,
		.alg_defaults	= &ccp_sm4_defaults,
	},
	{
		.mode		= CCP_SM4_ALG_MODE_ECB_HS,
		.version	= CCP_VERSION(5, 0),
		.name		= "ecb(sm4)",
		.driver_name	= "ecb-sm4-hs-ccp",
		.blocksize	= SM4_BLOCK_SIZE,
		.ivsize		= 0,
		.alg_defaults	= &ccp_sm4_defaults,
	},
	{
		.mode		= CCP_SM4_ALG_MODE_CBC,
		.version	= CCP_VERSION(5, 0),
		.name		= "cbc(sm4)",
		.driver_name	= "cbc-sm4-ccp",
		.blocksize	= SM4_BLOCK_SIZE,
		.ivsize		= SM4_BLOCK_SIZE,
		.alg_defaults	= &ccp_sm4_defaults,
	},
	{
		.mode		= CCP_SM4_ALG_MODE_CBC_HS,
		.version	= CCP_VERSION(5, 0),
		.name		= "cbc(sm4)",
		.driver_name	= "cbc-sm4-hs-ccp",
		.blocksize	= SM4_BLOCK_SIZE,
		.ivsize		= SM4_BLOCK_SIZE,
		.alg_defaults	= &ccp_sm4_defaults,
	},
	{
		.mode		= CCP_SM4_ALG_MODE_OFB,
		.version	= CCP_VERSION(5, 0),
		.name		= "ofb(sm4)",
		.driver_name	= "ofb-sm4-ccp",
		.blocksize	= SM4_BLOCK_SIZE,
		.ivsize		= SM4_BLOCK_SIZE,
		.alg_defaults	= &ccp_sm4_defaults,
	},
	{
		.mode		= CCP_SM4_ALG_MODE_CFB,
		.version	= CCP_VERSION(5, 0),
		.name		= "cfb(sm4)",
		.driver_name	= "cfb-sm4-ccp",
		.blocksize	= SM4_BLOCK_SIZE,
		.ivsize		= SM4_BLOCK_SIZE,
		.alg_defaults	= &ccp_sm4_defaults,
	},
	{
		.mode		= CCP_SM4_ALG_MODE_CTR,
		.version	= CCP_VERSION(5, 0),
		.name		= "ctr(sm4)",
		.driver_name	= "ctr-sm4-ccp",
		.blocksize	= 1,
		.ivsize		= SM4_BLOCK_SIZE,
		.alg_defaults	= &ccp_sm4_defaults,
	},
};

static int ccp_register_sm4_alg(struct list_head *head,
				const struct ccp_sm4_def *def)
{
	struct ccp_crypto_ablkcipher_alg *ccp_alg;
	struct crypto_alg *alg;
	int ret;

	ccp_alg = kzalloc(sizeof(*ccp_alg), GFP_KERNEL);
	if (!ccp_alg)
		return -ENOMEM;

	INIT_LIST_HEAD(&ccp_alg->entry);

	ccp_alg->mode = def->mode;

	/* Copy the defaults and override as necessary */
	alg = &ccp_alg->alg;
	*alg = *def->alg_defaults;
	snprintf(alg->cra_name, CRYPTO_MAX_ALG_NAME, "%s", def->name);
	snprintf(alg->cra_driver_name, CRYPTO_MAX_ALG_NAME, "%s",
		 def->driver_name);
	alg->cra_blocksize = def->blocksize;
	alg->cra_ablkcipher.ivsize = def->ivsize;

	ret = crypto_register_alg(alg);
	if (ret) {
		pr_err("%s ablkcipher algorithm registration error (%d)\n",
		       alg->cra_name, ret);
		kfree(ccp_alg);
		return ret;
	}

	list_add(&ccp_alg->entry, head);

	return 0;
}

int ccp_register_sm4_algs(struct list_head *head)
{
	int i, ret;
	unsigned int ccpversion = ccp_version();

	for (i = 0; i < ARRAY_SIZE(sm4_algs); i++) {
		if (sm4_algs[i].version > ccpversion)
			continue;
		ret = ccp_register_sm4_alg(head, &sm4_algs[i]);
		if (ret)
			return ret;
	}

	return 0;
}

