/*
 * SM2 key extract helper
 *
 * Copyright (C) 2019 Chengdu Haiguang Integrated Circuit Design Co., Ltd.
 *
 * Author: Panyang Du <dupanyang@hygon.cn>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 */
#include <linux/module.h>
#include <linux/export.h>
#include <linux/err.h>
#include "sm2.h"
#include "sm2pubkey.asn1.h"
#include "sm2privkey.asn1.h"

int sm2_get_d(void *context, size_t hdrlen, unsigned char tag,
	      const void *value, size_t vlen)
{
	struct sm2_key *key = context;

	/* invalid key provided */
	if (!value || !vlen)
		return -EINVAL;

	key->d = value;
	key->d_sz = vlen;

	return 0;
}

int sm2_get_q(void *context, size_t hdrlen, unsigned char tag,
	      const void *value, size_t vlen)
{
	struct sm2_key *key = context;
	const u8 *p = (const u8 *)value;

	/* invalid key provided */
	if (!value || !vlen)
		return -EINVAL;

	if (*p != 0x00)
		return -EINVAL;

	key->q = p + 1;
	key->q_sz = vlen - 1;

	return 0;
}

int sm2_check_ecpubkey(void *context, size_t hdrlen, unsigned char tag,
		const void *value, size_t vlen)
{
	/* 1.2.840.10045.2.1 : ANSI X9.62 public key type */
	const u8 pubkey_type[7] = {0x2a, 0x86, 0x48, 0xce, 0x3d, 0x02, 0x01};
	const u8 *p = (const u8 *)value;
	int ret;

	if (!value || vlen != 7)
		return -EINVAL;

	ret = memcmp(pubkey_type, p, vlen);
	if (ret)
		return -EINVAL;

	return 0;
}

int sm2_check_sm2ecc(void *context, size_t hdrlen, unsigned char tag,
		const void *value, size_t vlen)
{
	/* 1.2.156.10197.1.301 : China GM Standards Committee */
	const u8 gm_stands_val[8] = {0x2a, 0x81, 0x1c, 0xcf, 0x55,
					0x01, 0x82, 0x2d};
	const u8 *p = (const u8 *)value;
	int ret;

	if (!value || vlen != 8)
		return -EINVAL;

	ret = memcmp(gm_stands_val, p, vlen);
	if (ret)
		return -EINVAL;

	return 0;
}

/**
* sm2_parse_pub_key() - decodes the BER encoded buffer and stores in the
*		       provided struct sm2_key, pointers to the raw key as is,
*		       so that the caller can copy it, etc.
*
* @sm2_key:    struct sm2_key key representation
* @key:	       key in BER format
* @key_len:    length of key
*
* Return:      0 on success or error code in case of error
*/
int sm2_parse_pub_key(struct sm2_key *sm2_key, const void *key,
		    unsigned int key_len)
{
      return asn1_ber_decoder(&sm2pubkey_decoder, sm2_key, key, key_len);
}
EXPORT_SYMBOL_GPL(sm2_parse_pub_key);

/**
* sm2_parse_priv_key() - decodes the BER encoded buffer and stores in the
*			provided struct sm2_key, pointers to the raw key
*			as is, so that the caller can copy it, etc.
*
* @sm2_key:    struct rsa_key key representation
* @key:	      key in BER format
* @key_len:    length of key
*
* Return:      0 on success or error code in case of error
*/
int sm2_parse_priv_key(struct sm2_key *sm2_key, const void *key,
		     unsigned int key_len)
{
      return asn1_ber_decoder(&sm2privkey_decoder, sm2_key, key, key_len);
}
EXPORT_SYMBOL_GPL(sm2_parse_priv_key);

